# This file is part of allegedb, an object-relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""WindowDict, the core data structure used by allegedb's caching system.

It resembles a dictionary, more specifically a defaultdict-like where retrieving
a key that isn't set will get the highest set key that is lower than the key
you asked for (and thus, keys must be orderable). It is optimized for retrieval
of the same key and neighboring ones repeatedly and in sequence.

"""

from abc import abstractmethod, ABC
from collections import deque
from collections.abc import (
	Mapping,
	MutableMapping,
	KeysView,
	ItemsView,
	ValuesView,
)
from itertools import chain
from operator import itemgetter, lt, le
from threading import RLock
from typing import (
	Union,
	Callable,
	Dict,
	List,
	Tuple,
	Any,
	Iterable,
	Set,
	Optional,
)
from enum import Enum

get0 = itemgetter(0)
get1 = itemgetter(1)


class Direction(Enum):
	FORWARD = "forward"
	BACKWARD = "backward"


def update_window(
	turn_from: int,
	tick_from: int,
	turn_to: int,
	tick_to: int,
	updfun: Callable,
	branchd: Dict[int, List[tuple]],
):
	"""Iterate over some time in ``branchd``, call ``updfun`` on the values"""
	if turn_from in branchd:
		# Not including the exact tick you started from,
		# because deltas are *changes*
		for past_state in branchd[turn_from][tick_from + 1 :]:
			updfun(*past_state)
	for midturn in range(turn_from + 1, turn_to):
		if midturn in branchd:
			for past_state in branchd[midturn][:]:
				updfun(*past_state)
	if turn_to in branchd:
		for past_state in branchd[turn_to][: tick_to + 1]:
			updfun(*past_state)


def update_backward_window(
	turn_from: int,
	tick_from: int,
	turn_to: int,
	tick_to: int,
	updfun: Callable,
	branchd: Dict[int, List[tuple]],
):
	"""Iterate backward over time in ``branchd``, call ``updfun`` on the values"""
	if turn_from in branchd:
		for future_state in reversed(branchd[turn_from][: tick_from + 1]):
			updfun(*future_state)
	for midturn in range(turn_from - 1, turn_to, -1):
		if midturn in branchd:
			for future_state in reversed(branchd[midturn][:]):
				updfun(*future_state)
	if turn_to in branchd:
		for future_state in reversed(branchd[turn_to][tick_to:]):
			updfun(*future_state)


class HistoricKeyError(KeyError):
	"""Distinguishes deleted keys from those that were never set"""

	def __init__(self, *args, deleted=False):
		super().__init__(*args)
		self.deleted = deleted


class WindowDictKeysView(ABC, KeysView):
	"""Look through all the keys a WindowDict contains."""

	_mapping: "WindowDict"

	def __contains__(self, rev: int):
		with self._mapping._lock:
			return rev in self._mapping._keys

	def __iter__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if past:
				yield from map(get0, past)
			if future:
				yield from map(get0, reversed(future))


class WindowDictItemsView(ABC, ItemsView):
	"""Look through everything a WindowDict contains."""

	_mapping: "WindowDict"

	def __contains__(self, item: Tuple[int, Any]):
		with self._mapping._lock:
			return item in self._mapping._past or item in self._mapping._future

	def __iter__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if past:
				yield from past
			if future:
				yield from future


class WindowDictPastFutureKeysView(ABC, KeysView):
	"""View on a WindowDict's keys relative to last lookup"""

	_mapping: Union["WindowDictPastView", "WindowDictFutureView"]

	def __iter__(self):
		with self._mapping.lock:
			yield from map(get0, reversed(self._mapping.stack))

	def __contains__(self, item: int):
		with self._mapping.lock:
			return item in self._mapping and item in map(
				get0, self._mapping.stack
			)


class WindowDictPastFutureItemsView(ABC, ItemsView):
	_mapping: Union["WindowDictPastView", "WindowDictFutureView"]

	@staticmethod
	@abstractmethod
	def _out_of_range(item: tuple, stack: list):
		pass

	def __iter__(self):
		with self._mapping.lock:
			yield from reversed(self._mapping.stack)

	def __contains__(self, item: Tuple[int, Any]):
		with self._mapping.lock:
			if self._out_of_range(item, self._mapping.stack):
				return False
			i0, i1 = item
			for j0, j1 in self._mapping.stack:
				if i0 == j0:
					return i1 == j1
			return False


class WindowDictPastItemsView(WindowDictPastFutureItemsView):
	@staticmethod
	def _out_of_range(item: Tuple[int, Any], stack: List[Tuple[int, Any]]):
		return item[0] < stack[0][0] or item[0] > stack[-1][0]


class WindowDictFutureItemsView(WindowDictPastFutureItemsView):
	"""View on a WindowDict's future items relative to last lookup"""

	@staticmethod
	def _out_of_range(item: Tuple[int, Any], stack: List[Tuple[int, Any]]):
		return item[0] < stack[-1][0] or item[0] > stack[0][0]


class WindowDictPastFutureValuesView(ABC, ValuesView):
	"""Abstract class for views on the past or future values of a WindowDict"""

	_mapping: Union["WindowDictPastView", "WindowDictFutureView"]

	def __iter__(self):
		with self._mapping.lock:
			yield from map(get1, reversed(self._mapping.stack))

	def __contains__(self, item: Any):
		with self._mapping.lock:
			return item in map(get1, self._mapping.stack)


class WindowDictValuesView(ABC, ValuesView):
	"""Look through all the values that a WindowDict contains."""

	_mapping: "WindowDict"

	def __contains__(self, value: Any):
		with self._mapping._lock:
			return value in map(get1, self._mapping._past) or value in map(
				get1, self._mapping._future
			)

	def __iter__(self):
		with self._mapping._lock:
			past = self._mapping._past
			future = self._mapping._future
			if past:
				yield from map(get1, past)
			if future:
				yield from map(get1, future)


class WindowDictPastFutureView(ABC, Mapping):
	"""Abstract class for historical views on WindowDict"""

	__slots__ = ("stack", "lock")
	stack: List[Tuple[int, Any]]

	def __init__(self, stack: List[Tuple[int, Any]], lock: RLock) -> None:
		self.stack = stack
		self.lock = lock

	def __len__(self) -> int:
		with self.lock:
			stack = self.stack
			if not stack:
				return 0
			return len(stack)


class WindowDictPastView(WindowDictPastFutureView):
	"""Read-only mapping of just the past of a WindowDict"""

	def __iter__(self) -> Iterable[int]:
		with self.lock:
			stack = self.stack
			return map(get0, reversed(stack))

	def __getitem__(self, key: int) -> Any:
		with self.lock:
			stack = self.stack
			if not stack or key < stack[0][0] or key > stack[-1][0]:
				raise KeyError
			for rev, value in stack:
				if rev == key:
					return value
			raise KeyError

	def keys(self) -> WindowDictPastFutureKeysView:
		return WindowDictPastFutureKeysView(self)

	def items(self) -> WindowDictPastItemsView:
		return WindowDictPastItemsView(self)

	def values(self) -> WindowDictPastFutureValuesView:
		return WindowDictPastFutureValuesView(self)


class WindowDictFutureView(WindowDictPastFutureView):
	"""Read-only mapping of just the future of a WindowDict"""

	def __iter__(self):
		with self.lock:
			stack = self.stack
			return map(get0, reversed(stack))

	def __getitem__(self, key: int):
		with self.lock:
			stack = self.stack
			if not stack or key < stack[-1][0] or key > stack[0][0]:
				raise KeyError
			for rev, value in stack:
				if rev == key:
					return value
			raise KeyError("No such revision", key)

	def keys(self) -> WindowDictPastFutureKeysView:
		return WindowDictPastFutureKeysView(self)

	def items(self) -> WindowDictFutureItemsView:
		return WindowDictFutureItemsView(self)

	def values(self) -> WindowDictPastFutureValuesView:
		return WindowDictPastFutureValuesView(self)


class WindowDictSlice:
	"""A slice of history in which the start is earlier than the stop"""

	__slots__ = ["dic", "slic"]
	dic: "WindowDict"
	slic: slice

	def __init__(self, dic: "WindowDict", slic: slice):
		self.dic = dic
		self.slic = slic

	def __reversed__(self) -> Iterable[Any]:
		return iter(WindowDictReverseSlice(self.dic, self.slic))

	def __iter__(self):
		dic = self.dic
		with dic._lock:
			if not dic:
				return
			slic = self.slic
			if slic.step is not None:
				for i in range(
					slic.start or dic.beginning,
					slic.stop or dic.end + 1,
					slic.step,
				):
					dic._seek(i)
					yield dic._past[-1][1]
				return
			if slic.start is None and slic.stop is None:
				yield from map(get1, dic._past)
				yield from map(get1, reversed(dic._future))
				return
			if slic.start is not None and slic.stop is not None:
				if slic.stop == slic.start:
					try:
						yield dic[slic.stop]
					except HistoricKeyError:
						pass
					return
				past = dic._past
				future = dic._future
				if slic.start < slic.stop:
					left, right = slic.start, slic.stop
					dic._seek(right)
					if not past:
						return
					if past[-1][0] == right:
						future.append(past.pop())
					cmp = lt
				else:
					left, right = slic.stop, slic.start
					dic._seek(right)
					cmp = le
				if not past:
					return
				it = iter(past)
				p0, p1 = next(it)
				while cmp(p0, left):
					try:
						p0, p1 = next(it)
					except StopIteration:
						return
				else:
					yield p1
				yield from map(get1, it)
			elif slic.start is None:
				stac = dic._past + list(reversed(dic._future))
				while stac and stac[-1][0] >= slic.stop:
					stac.pop()
				yield from map(get1, stac)
				return
			else:  # slic.stop is None
				if not dic._past and not dic._future:
					return
				chan = chain(dic._past, reversed(dic._future))
				nxt = next(chan)
				while nxt[0] < slic.start:
					try:
						nxt = next(chan)
					except StopIteration:
						return
				yield get1(nxt)
				yield from map(get1, chan)


class WindowDictReverseSlice:
	"""A slice of history in which the start is later than the stop"""

	__slots__ = ["dict", "slice"]

	def __init__(self, dict: "WindowDict", slic: slice):
		self.dict = dict
		self.slice = slic

	def __reversed__(self):
		return iter(WindowDictSlice(self.dict, self.slice))

	def __iter__(self):
		dic = self.dict
		with dic._lock:
			if not dic:
				return
			slic = self.slice
			if slic.step is not None:
				for i in range(
					slic.start or dic.end,
					slic.stop or dic.beginning,
					slic.step,
				):
					dic._seek(i)
					yield dic._past[-1][1]
				return
			if slic.start is None and slic.stop is None:
				yield from map(get1, dic._future)
				yield from map(get1, reversed(dic._past))
				return
			if slic.start is not None and slic.stop is not None:
				if slic.start == slic.stop:
					dic._seek(slic.stop)
					yield dic._past[-1][1]
					return
				if slic.start < slic.stop:
					left, right = slic.start, slic.stop
					dic._seek(right)
					it = reversed(dic._past)
					next(it)
					cmp = lt
				else:
					left, right = slic.stop, slic.start
					dic._seek(right)
					it = reversed(dic._past)
					cmp = le
				for frev, fv in it:
					if cmp(frev, left):
						return
					yield fv
			elif slic.start is None:
				stac = dic._past + list(reversed(dic._future))
				while stac and stac[-1][0] >= slic.stop:
					stac.pop()
				yield from map(get1, reversed(stac))
			else:  # slic.stop is None
				stac = deque(dic._past)
				stac.extend(reversed(dic._future))
				while stac and stac[0][0] < slic.start:
					stac.popleft()
				yield from map(get1, reversed(stac))


class WindowDict(MutableMapping):
	"""A dict that keeps every value that a variable has had over time.

	Look up a revision number in this dict, and it will give you the
	effective value as of that revision. Keys should always be
	revision numbers.

	Optimized for the cases where you look up the same revision
	repeatedly, or its neighbors.

	This supports slice notation to get all values in a given
	time-frame. If you do not supply a step, you'll just get the
	values, with no indication of when they're from exactly --
	so explicitly supply a step of 1 to get the value at each point in
	the slice, or use the ``future`` and ``past`` methods to get read-only
	mappings of data relative to a particular revision.

	Unlike slices of eg. lists, you can slice with a start greater than the stop
	even if you don't supply a step. That will get you values in reverse order.

	"""

	__slots__ = ("_future", "_past", "_keys", "_last", "_lock")

	_past: List[Tuple[int, Any]]
	_future: List[Tuple[int, Any]]
	_keys: Set[int]
	_last: Optional[int]

	@property
	def beginning(self) -> Optional[int]:
		with self._lock:
			if not self._past:
				if not self._future:
					return None
				return self._future[-1][0]
			return self._past[0][0]

	@property
	def end(self) -> Optional[int]:
		with self._lock:
			if not self._future:
				if not self._past:
					return None
				return self._past[-1][0]
			return self._future[0][0]

	def future(self, rev: int = None) -> WindowDictFutureView:
		"""Return a Mapping of items after the given revision.

		Default revision is the last one looked up.

		"""
		if rev is not None:
			with self._lock:
				self._seek(rev)
		return WindowDictFutureView(self._future, self._lock)

	def past(self, rev: int = None) -> WindowDictPastView:
		"""Return a Mapping of items at or before the given revision.

		Default revision is the last one looked up.

		"""
		if rev is not None:
			with self._lock:
				self._seek(rev)
		return WindowDictPastView(self._past, self._lock)

	def search(self, rev: int) -> Any:
		"""Alternative access for far-away revisions

		This uses a binary search, which is faster in the case of random
		access, but not in the case of fast-forward and rewind, which are
		more common in time travel.

		This arranges the cache to optimize retrieval of the same and
		nearby revisions, same as normal lookups.

		"""

		def recurse(revs: List[Tuple[int, Any]]) -> Any:
			if len(revs) < 1:
				raise HistoricKeyError(
					"No data ever for revision", rev, deleted=False
				)
			elif len(revs) == 1:
				if revs[0][0] <= rev:
					return revs[0]
				raise HistoricKeyError(
					"Can't retrieve revision", rev, deleted=True
				)
			pivot = len(revs) // 2
			before = revs[:pivot]
			after = revs[pivot:]
			assert before and after
			if rev < after[0][0]:
				if rev > before[-1][0]:
					return before[-1]
				return recurse(before)
			elif rev == after[0][0]:
				return after[0]
			else:
				return recurse(after)

		with self._lock:
			revs = self._past + list(reversed(self._future))
			if len(revs) == 1:
				result_rev, result = revs[0]
				if rev < result_rev:
					raise HistoricKeyError(
						"No data ever for revision", rev, deleted=False
					)
			else:
				result_rev, result = recurse(revs)
			i = revs.index((result_rev, result)) + 1
			self._past = revs[:i]
			self._future = list(reversed(revs[i:]))
			self._last = rev
			return result

	def _seek(self, rev: int) -> None:
		"""Arrange the caches to help look up the given revision."""
		if rev == self._last:
			return
		past = self._past
		future = self._future
		if future:
			appender = past.append
			popper = future.pop
			future_start = future[-1][0]
			while future_start <= rev:
				appender(popper())
				if future:
					future_start = future[-1][0]
				else:
					break
		if past:
			popper = past.pop
			appender = future.append
			past_end = past[-1][0]
			while past_end > rev:
				appender(popper())
				if past:
					past_end = past[-1][0]
				else:
					break
		self._last = rev

	def rev_gettable(self, rev: int) -> bool:
		beg = self.beginning
		if beg is None:
			return False
		return rev >= beg

	def rev_before(self, rev: int, search=False):
		"""Return the latest past rev on which the value changed."""
		with self._lock:
			if search:
				self.search(rev)
			else:
				self._seek(rev)
			if self._past:
				return self._past[-1][0]

	def rev_after(self, rev: int, search=False):
		"""Return the earliest future rev on which the value will change."""
		with self._lock:
			if search:
				self.search(rev)
			else:
				self._seek(rev)
			if self._future:
				return self._future[-1][0]

	def initial(self) -> Any:
		"""Return the earliest value we have"""
		with self._lock:
			if self._past:
				return self._past[0][1]
			if self._future:
				return self._future[-1][1]
			raise KeyError("No data")

	def final(self) -> Any:
		"""Return the latest value we have"""
		with self._lock:
			if self._future:
				return self._future[0][1]
			if self._past:
				return self._past[-1][1]
			raise KeyError("No data")

	def truncate(
		self, rev: int, direction: Direction = "forward", search=False
	) -> None:
		"""Delete everything after the given revision, exclusive.

		With direction='backward', delete everything before the revision,
		exclusive, instead.

		"""
		with self._lock:
			if search:
				self.search(rev)
			else:
				self._seek(rev)
			if direction == "forward":
				self._keys.difference_update(map(get0, self._future))
				self._future = []
			elif direction == "backward":
				if not self._past:
					return
				if self._past[-1][0] == rev:
					self._keys.difference_update(map(get0, self._past[:-1]))
					self._past = [self._past[-1]]
				else:
					self._keys.difference_update(map(get0, self._past))
					self._past = []
			else:
				raise ValueError("Need direction 'forward' or 'backward'")

	def keys(self) -> WindowDictKeysView:
		return WindowDictKeysView(self)

	def items(self) -> WindowDictItemsView:
		return WindowDictItemsView(self)

	def values(self) -> WindowDictValuesView:
		return WindowDictValuesView(self)

	def __bool__(self) -> bool:
		return bool(self._keys)

	def copy(self):
		with self._lock:
			empty = WindowDict.__new__(WindowDict)
			empty._past = self._past.copy()
			empty._future = self._future.copy()
			empty._keys = self._keys.copy()
			empty._last = self._last
			return empty

	def __init__(
		self, data: Union[List[Tuple[int, Any]], Dict[int, Any]] = None
	) -> None:
		self._lock = RLock()
		with self._lock:
			if not data:
				self._past = []
			elif isinstance(data, Mapping):
				self._past = list(data.items())
			else:
				# assume it's an orderable sequence of pairs
				self._past = list(data)
			self._past.sort()
			self._future = []
			self._keys = set(map(get0, self._past))
			self._last = None

	def __iter__(self) -> Iterable[Any]:
		if not self:
			return
		if self._past:
			yield from map(get0, self._past)
		if self._future:
			yield from map(get0, self._future)

	def __contains__(self, item: int) -> bool:
		return item in self._keys

	def __len__(self) -> int:
		return len(self._keys)

	def __getitem__(self, rev: int) -> Any:
		if isinstance(rev, slice):
			if None not in (rev.start, rev.stop) and rev.start > rev.stop:
				return WindowDictReverseSlice(self, rev)
			return WindowDictSlice(self, rev)
		with self._lock:
			self._seek(rev)
			past = self._past
			if not past:
				raise HistoricKeyError(
					"Revision {} is before the start of history".format(rev)
				)
			return past[-1][1]

	def __setitem__(self, rev: int, v: Any) -> None:
		self.set_item(rev, v)

	def set_item(self, rev: int, v: Any, search=False) -> None:
		past = self._past
		with self._lock:
			if past or self._future:
				if search:
					self.search(rev)
				else:
					self._seek(rev)
				if past:
					if past[-1][0] == rev:
						past[-1] = (rev, v)
					else:
						past.append((rev, v))
				else:
					past.append((rev, v))
			else:
				past.append((rev, v))
			self._keys.add(rev)

	def __delitem__(self, rev: int) -> None:
		self.del_item(rev)

	def del_item(self, rev: int, search=False) -> None:
		# Not checking for rev's presence at the beginning because
		# to do so would likely require iterating thru history,
		# which I have to do anyway in deleting.
		# But handle degenerate case.
		if not self:
			raise HistoricKeyError("Tried to delete from an empty WindowDict")
		if self.beginning is None:
			if self.end is not None and rev > self.end:
				raise HistoricKeyError(
					"Rev outside of history: {}".format(rev)
				)
		elif self.end is None:
			if self.beginning is not None and rev < self.beginning:
				raise HistoricKeyError(
					"Rev outside of history: {}".format(rev)
				)
		elif not self.beginning <= rev <= self.end:
			raise HistoricKeyError("Rev outside of history: {}".format(rev))
		with self._lock:
			if search:
				self.search(rev)
			else:
				self._seek(rev)
			past = self._past
			if not past or past[-1][0] != rev:
				raise HistoricKeyError("Rev not present: {}".format(rev))
			del past[-1]
			self._keys.remove(rev)

	def __repr__(self) -> str:
		me = {}
		if self._past:
			me.update(self._past)
		if self._future:
			me.update(self._future)
		return "{}({})".format(self.__class__.__name__, me)


class FuturistWindowDict(WindowDict):
	"""A WindowDict that does not let you rewrite the past."""

	__slots__ = (
		"_future",
		"_past",
	)
	_future: List[Tuple[int, Any]]
	_past: List[Tuple[int, Any]]

	def __setitem__(self, rev: int, v: Any) -> None:
		if hasattr(v, "unwrap") and not hasattr(v, "no_unwrap"):
			v = v.unwrap()
		with self._lock:
			self._seek(rev)
			past = self._past
			future = self._future
			if future:
				raise HistoricKeyError(
					"Already have some history after {}".format(rev)
				)
			if not past:
				past.append((rev, v))
			elif rev > past[-1][0]:
				past.append((rev, v))
			elif rev == past[-1][0]:
				past[-1] = (rev, v)
			else:
				raise HistoricKeyError(
					"Already have some history after {} "
					"(and my seek function is broken?)".format(rev)
				)
			self._keys.add(rev)


class TurnDict(FuturistWindowDict):
	__slots__ = ("_future", "_past")
	_future: List[Tuple[int, Any]]
	_past: List[Tuple[int, Any]]
	cls = FuturistWindowDict

	def __setitem__(self, turn: int, value: Any) -> None:
		if type(value) is not FuturistWindowDict:
			value = FuturistWindowDict(value)
		FuturistWindowDict.__setitem__(self, turn, value)


class EntikeyWindowDict(WindowDict):
	__slots__ = ("_past", "_future", "entikeys")

	def __init__(
		self, data: Union[List[Tuple[int, Any]], Dict[int, Any]] = None
	) -> None:
		if data:
			if hasattr(data, "values") and callable(data.values):
				self.entikeys = {value[:-2] for value in data.values()}
			else:
				self.entikeys = {value[:-2] for value in data}
		else:
			self.entikeys = set()
		super().__init__(data)

	def __setitem__(self, rev: int, v: tuple) -> None:
		self.entikeys.add(v[:-2])
		super().__setitem__(rev, v)

	def __delitem__(self, rev: int) -> None:
		entikey = self[rev][:-2]
		super().__delitem__(rev)
		for tup in self.values():
			if tup[:-2] == entikey:
				return
		self.entikeys.remove(entikey)


class SettingsTurnDict(WindowDict):
	"""A WindowDict that contains a span of time, indexed as turns and ticks

	Each turn is a series of ticks. Once a value is set at some turn and tick,
	it's in effect at every tick in the turn after that one, and every
	further turn.

	"""

	__slots__ = ("_future", "_past")
	_future: List[Tuple[int, Any]]
	_past: List[Tuple[int, Any]]
	cls = WindowDict

	def __setitem__(self, turn: int, value: Any) -> None:
		if not isinstance(value, self.cls):
			value = self.cls(value)
		WindowDict.__setitem__(self, turn, value)

	def retrieve(self, turn: int, tick: int) -> Any:
		"""Retrieve the value that was in effect at this turn and tick

		Whether or not it was *set* at this turn and tick

		"""
		if turn in self and self[turn].rev_gettable(tick):
			return self[turn][tick]
		elif self.rev_gettable(turn - 1):
			return self[turn - 1].final()
		raise KeyError(f"Can't retrieve turn {turn}, tick {tick}")

	def retrieve_exact(self, turn: int, tick: int) -> Any:
		"""Retrieve the value only if it was set at this exact turn and tick"""
		if turn not in self:
			raise KeyError(f"No data in turn {turn}")
		if tick not in self[turn]:
			raise KeyError(f"No data for tick {tick} in turn {turn}")
		return self[turn][tick]

	def store_at(self, turn: int, tick: int, value: Any) -> None:
		"""Set a value at a time, creating the turn if needed"""
		if turn in self:
			self[turn][tick] = value
		else:
			self[turn] = {tick: value}


class EntikeySettingsTurnDict(SettingsTurnDict):
	cls = EntikeyWindowDict



from time import monotonic

import networkx as nx

from LiSE import Engine
from LiSE.proxy import EngineProcessManager


def test_follow_path(tmp_path):
	big_grid = nx.grid_2d_graph(100, 100)
	big_grid.add_node("them", location=(0, 0))
	straightly = nx.shortest_path(big_grid, (0, 0), (99, 99))
	with Engine(tmp_path) as eng:
		eng.add_character("grid", big_grid)
	with EngineProcessManager(tmp_path) as prox:
		them = prox.character["grid"].thing["them"]
		start = monotonic()
		them.follow_path(straightly)
		elapsed = monotonic() - start
		assert (
			elapsed < 0.15
		), f"Took too long to follow a path of length {len(straightly)}: {elapsed:.2} seconds"



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from functools import partial
from operator import sub, or_

from .allegedb import Key
from .allegedb.cache import (
	Cache,
	PickyDefaultDict,
	StructuredDefaultDict,
	TurnDict,
	WindowDict,
	HistoricKeyError,
	EntitylessCache,
)
from .allegedb.window import EntikeySettingsTurnDict, SettingsTurnDict
from .util import singleton_get, sort_set
from collections import OrderedDict


class InitializedCache(Cache):
	__slots__ = ()

	def _store_journal(self, *args):
		entity, key, branch, turn, tick, value = args[-6:]
		parent = args[:-6]
		settings_turns = self.settings[branch]
		presettings_turns = self.presettings[branch]
		try:
			prev = self.retrieve(*args[:-1])
		except KeyError:
			prev = None
		if prev == value:
			return  # not much point reporting on a non-change in a diff
		if turn in settings_turns or turn in settings_turns.future():
			assert (
				turn in presettings_turns or turn in presettings_turns.future()
			)
			setticks = settings_turns[turn]
			presetticks = presettings_turns[turn]
			presetticks[tick] = parent + (entity, key, prev)
			setticks[tick] = parent + (entity, key, value)
		else:
			presettings_turns[turn] = {tick: parent + (entity, key, prev)}
			settings_turns[turn] = {tick: parent + (entity, key, value)}


class InitializedEntitylessCache(EntitylessCache, InitializedCache):
	__slots__ = ()


class UnitnessCache(Cache):
	"""A cache for remembering when a node is a unit of a character."""

	__slots__ = (
		"user_order",
		"user_shallow",
		"graphs",
		"graph_units",
		"char_units",
		"solo_unit",
		"unique_unit",
		"unique_graph",
		"users",
	)

	def __init__(self, engine):
		Cache.__init__(self, engine)
		self.user_order = StructuredDefaultDict(3, TurnDict)
		self.user_shallow = PickyDefaultDict(TurnDict)
		self.graphs = StructuredDefaultDict(1, TurnDict)
		self.graph_units = StructuredDefaultDict(1, TurnDict)
		self.char_units = StructuredDefaultDict(1, TurnDict)
		self.solo_unit = StructuredDefaultDict(1, TurnDict)
		self.unique_unit = StructuredDefaultDict(1, TurnDict)
		self.unique_graph = StructuredDefaultDict(1, TurnDict)
		self.users = StructuredDefaultDict(1, TurnDict)

	def store(
		self,
		character,
		graph,
		node,
		branch,
		turn,
		tick,
		is_unit,
		*,
		planning=None,
		forward=None,
		loading=False,
		contra=True,
	):
		is_unit = True if is_unit else None
		super().store(
			character,
			graph,
			node,
			branch,
			turn,
			tick,
			is_unit,
			planning=planning,
			forward=forward,
			loading=loading,
			contra=contra,
		)
		userturns = self.user_order[graph][node][character][branch]
		if turn in userturns:
			userturns[turn][tick] = is_unit
		else:
			userturns[turn] = {tick: is_unit}
		usershal = self.user_shallow[(graph, node, character, branch)]
		if turn in usershal:
			usershal[turn][tick] = is_unit
		else:
			usershal[turn] = {tick: is_unit}
		charavs = self.char_units[character]
		graphavs = self.graph_units[(character, graph)]
		graphs = self.graphs[character]
		uniqgraph = self.unique_graph[character][branch]
		soloav = self.solo_unit[(character, graph)][branch]
		uniqav = self.unique_unit[character][branch]
		users = self.users[graph, node]

		def add_or_discard_something(op, cache, what):
			for b, r, t in self.db._iter_parent_btt(branch, turn, tick):
				if b in cache:
					cb = cache[b]
					if r in cb and cb[r].rev_gettable(t):
						old_cache = cb[r][t]
						break
					elif cb.rev_gettable(r - 1):
						old_cache = cb[r - 1].final()
						break
			else:
				old_cache = frozenset()
			new_cache = op(old_cache, frozenset((what,)))
			if turn in cache[branch]:
				cache[branch][turn][tick] = new_cache
			elif branch in cache:
				cache[branch][turn] = {tick: new_cache}
			else:
				cache[branch] = {turn: {tick: new_cache}}

		add_something = partial(add_or_discard_something, or_)
		discard_something = partial(add_or_discard_something, sub)

		if is_unit:
			add_something(graphavs, node)
			add_something(charavs, (graph, node))
			add_something(graphs, graph)
			add_something(users, character)
		else:
			discard_something(graphavs, node)
			discard_something(charavs, (graph, node))
			try:
				if not graphavs[turn][tick]:
					discard_something(graphs, graph)
			except HistoricKeyError:
				pass
			try:
				if not charavs[turn][tick]:
					discard_something(users, character)
			except HistoricKeyError:
				pass
		try:
			graphav = singleton_get(graphavs[turn][tick])
			if turn in soloav:
				soloav[turn][tick] = graphav
			else:
				soloav[turn] = {tick: graphav}
		except HistoricKeyError:
			pass
		try:
			charav = singleton_get(charavs[turn][tick])
			if turn in uniqav:
				uniqav[turn][tick] = charav
			else:
				uniqav[turn] = {tick: charav}
		except HistoricKeyError:
			pass
		try:
			if not graphavs[turn][tick]:
				if graph in graphs[turn][tick]:
					graphs[turn][tick].remove(graph)
					if len(graphs[turn][tick]) == 1:
						uniqgraph[turn][tick] = next(iter(graphs[turn][tick]))
					else:
						uniqgraph[turn][tick] = None
		except HistoricKeyError:
			pass
		if (
			turn in graphavs
			and tick in graphavs[turn]
			and len(graphavs[turn][tick]) != 1
		):
			if turn in soloav:
				soloav[turn][tick] = None
			else:
				soloav[turn] = {tick: None}
		else:
			if turn in soloav:
				soloav[turn][tick] = node
			else:
				soloav[turn] = {tick: None}
		if (
			turn in charavs
			and charavs[turn].rev_gettable(tick)
			and len(charavs[turn][tick]) != 1
		):
			if turn in uniqav:
				uniqav[turn][tick] = None
			else:
				uniqav[turn] = {tick: None}
		elif turn in uniqav:
			uniqav[turn][tick] = (graph, node)
		else:
			uniqav[turn] = {tick: (graph, node)}
		if (
			turn in graphs
			and graphs[turn].rev_gettable(tick)
			and len(graphs[turn][tick]) != 1
		):
			if turn in uniqgraph:
				uniqgraph[turn][tick] = None
			else:
				uniqgraph[turn] = {tick: None}
		elif turn in uniqgraph:
			uniqgraph[turn][tick] = graph
		else:
			uniqgraph[turn] = {tick: graph}

	def get_char_graph_avs(self, char, graph, branch, turn, tick):
		return (
			self._valcache_lookup(
				self.graph_units[(char, graph)], branch, turn, tick
			)
			or set()
		)

	def get_char_graph_solo_av(self, char, graph, branch, turn, tick):
		return self._valcache_lookup(
			self.solo_unit[(char, graph)], branch, turn, tick
		)

	def get_char_only_av(self, char, branch, turn, tick):
		return self._valcache_lookup(
			self.unique_unit[char], branch, turn, tick
		)

	def get_char_only_graph(self, char, branch, turn, tick):
		return self._valcache_lookup(
			self.unique_graph[char], branch, turn, tick
		)

	def get_char_graphs(self, char, branch, turn, tick):
		return (
			self._valcache_lookup(self.graphs[char], branch, turn, tick)
			or set()
		)

	def _slow_iter_character_avatars(
		self, character, branch, turn, tick, *, forward
	):
		for graph in self.iter_entities(
			character, branch, turn, tick, forward=forward
		):
			for node in self.iter_entities(
				character, graph, branch, turn, tick, forward=forward
			):
				yield graph, node

	def _slow_iter_users(self, graph, node, branch, turn, tick):
		if graph not in self.user_order:
			return
		for character in self.user_order[graph][node]:
			if (graph, node, character, branch) not in self.user_shallow:
				for b, t, tc in self.db._iter_parent_btt(branch, turn, tick):
					if b in self.user_order[graph][node][character]:
						isav = self.user_order[graph][node][character][b][t]
						# side effect!! bad!
						self.store(
							character,
							graph,
							node,
							branch,
							turn,
							tick,
							isav[tc],
						)
						break
				else:
					self.store(
						character, graph, node, branch, turn, tick, None
					)
			try:
				if self.user_shallow[(graph, node, character, branch)][turn][
					tick
				]:
					yield character
			except HistoricKeyError:
				continue


class RulesHandledCache(object):
	def __init__(self, engine):
		self.engine = engine
		self.handled = {}
		self.handled_deep = StructuredDefaultDict(1, type=WindowDict)
		self.unhandled = {}

	def get_rulebook(self, *args):
		raise NotImplementedError

	def iter_unhandled_rules(self, branch, turn, tick):
		raise NotImplementedError

	def store(self, *args, loading=False):
		entity = args[:-5]
		rulebook, rule, branch, turn, tick = args[-5:]
		self.handled.setdefault(entity + (rulebook, branch, turn), set()).add(
			rule
		)
		self.handled_deep[branch][turn][tick] = (entity, rulebook, rule)
		unhandl = (
			self.unhandled.setdefault(entity, {})
			.setdefault(rulebook, {})
			.setdefault(branch, {})
		)
		if turn not in unhandl:
			unhandl[turn] = list(
				self.unhandled_rulebook_rules(
					entity, rulebook, branch, turn, tick
				)
			)
		try:
			unhandl[turn].remove(rule)
		except ValueError:
			pass

	def retrieve(self, *args):
		return self.handled[args]

	def unhandled_rulebook_rules(self, *args):
		entity = args[:-4]
		rulebook, branch, turn, tick = args[-4:]
		unh = self.unhandled
		if entity in unh:
			ue = unh[entity]
			if rulebook in ue:
				uer = ue[rulebook]
				if branch in uer:
					uerb = uer[branch]
					if turn in uerb:
						return uerb[turn]
		rbc = self.engine._rulebooks_cache
		if not rbc.contains_key(rulebook, branch, turn, tick):
			return []
		rulebook_rules = rbc.retrieve(rulebook, branch, turn, tick)[0]
		handled_rules = self.handled.setdefault(
			entity + (rulebook, branch, turn), set()
		)
		return [rule for rule in rulebook_rules if rule not in handled_rules]


class CharacterRulesHandledCache(RulesHandledCache):
	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_rulebooks_cache.retrieve(
				character, branch, turn, tick
			)
		except KeyError:
			return character, "character"

	def iter_unhandled_rules(self, branch, turn, tick):
		for character in self.engine.character.keys():
			rb = self.get_rulebook(character, branch, turn, tick)
			try:
				prio = self.engine._rulebooks_cache.retrieve(
					rb, branch, turn, tick
				)[1]
			except KeyError:
				continue
			for rule in self.unhandled_rulebook_rules(
				character, rb, branch, turn, tick
			):
				yield prio, character, rb, rule


class UnitRulesHandledCache(RulesHandledCache):
	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._units_rulebooks_cache.retrieve(
				character, branch, turn, tick
			)
		except KeyError:
			return character, "unit"

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			rulebook = self.get_rulebook(character, branch, turn, tick)
			try:
				prio = self.engine._rulebooks_cache.retrieve(
					rulebook, branch, turn, tick
				)[1]
			except KeyError:
				continue
			charavm = charm[character].unit
			for graph in sort_set(charavm.keys()):
				for avatar in sort_set(charavm[graph].keys()):
					try:
						rules = self.unhandled_rulebook_rules(
							character,
							graph,
							avatar,
							rulebook,
							branch,
							turn,
							tick,
						)
					except KeyError:
						continue
					for rule in rules:
						yield prio, character, graph, avatar, rulebook, rule


class CharacterThingRulesHandledCache(RulesHandledCache):
	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_things_rulebooks_cache.retrieve(
				character, branch, turn, tick
			)
		except KeyError:
			return character, "character_thing"

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in charm.keys():
			rulebook = self.get_rulebook(character, branch, turn, tick)
			try:
				prio = self.engine._rulebooks_cache.retrieve(
					rulebook, branch, turn, tick
				)[1]
			except KeyError:
				continue
			for thing in charm[character].thing.keys():
				try:
					rules = self.unhandled_rulebook_rules(
						character, thing, rulebook, branch, turn, tick
					)
				except KeyError:
					continue
				for rule in rules:
					yield prio, character, thing, rulebook, rule


class CharacterPlaceRulesHandledCache(RulesHandledCache):
	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_places_rulebooks_cache.retrieve(
				character, branch, turn, tick
			)
		except KeyError:
			return character, "character_place"

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in charm.keys():
			rulebook = self.get_rulebook(character, branch, turn, tick)
			try:
				prio = self.engine._rulebooks_cache.retrieve(
					rulebook, branch, turn, tick
				)[1]
			except KeyError:
				continue
			for place in charm[character].place.keys():
				try:
					rules = self.unhandled_rulebook_rules(
						character, place, rulebook, branch, turn, tick
					)
				except KeyError:
					continue
				for rule in rules:
					yield prio, character, place, rulebook, rule


class CharacterPortalRulesHandledCache(RulesHandledCache):
	def get_rulebook(self, character, branch, turn, tick):
		try:
			return self.engine._characters_portals_rulebooks_cache.retrieve(
				character, branch, turn, tick
			)
		except KeyError:
			return character, "character_portal"

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		for character in sort_set(charm.keys()):
			try:
				rulebook = self.get_rulebook(character, branch, turn, tick)
			except KeyError:
				continue
			try:
				rulebook_rules, prio = self.engine._rulebooks_cache.retrieve(
					rulebook, branch, turn, tick
				)
			except KeyError:
				continue
			if not rulebook_rules:
				continue
			char = charm[character]
			charn = char.node
			charp = char.portal
			for orig in charp.keys():
				if orig not in charn:
					continue
				for dest in charp[orig].keys():
					if dest not in charn:
						continue
					try:
						rules = self.unhandled_rulebook_rules(
							character, orig, dest, rulebook, branch, turn, tick
						)
					except KeyError:
						continue
					for rule in rules:
						yield prio, character, orig, dest, rulebook, rule


class NodeRulesHandledCache(RulesHandledCache):
	def get_rulebook(self, character, node, branch, turn, tick):
		try:
			return self.engine._nodes_rulebooks_cache.retrieve(
				character, node, branch, turn, tick
			)
		except KeyError:
			return character, node

	def iter_unhandled_rules(self, branch, turn, tick):
		charm = self.engine.character
		nodes_with_rulebook_changed = {
			(character, node)
			for (
				character,
				node,
				_,
				_,
				_,
			) in self.engine._nodes_rulebooks_cache.shallowest
		}
		nodes_with_filled_default_rulebooks = {
			(character, node)
			for (_, (character, node)) in (
				{
					(None, (character, node))
					for character in charm
					for node in charm[character].node
				}
				& self.engine._rulebooks_cache.branches.keys()
			)
		}
		for character, node in (
			nodes_with_rulebook_changed | nodes_with_filled_default_rulebooks
		):
			try:
				rulebook = self.get_rulebook(
					character, node, branch, turn, tick
				)
				prio = self.engine._rulebooks_cache.retrieve(
					rulebook, branch, turn, tick
				)[1]
				rules = self.unhandled_rulebook_rules(
					character, node, rulebook, branch, turn, tick
				)
			except KeyError:
				continue
			for rule in rules:
				yield prio, character, node, rulebook, rule


class PortalRulesHandledCache(RulesHandledCache):
	def get_rulebook(self, character, orig, dest, branch, turn, tick):
		try:
			return self.engine._portals_rulebooks_cache.retrieve(
				character, orig, dest, branch, turn, tick
			)
		except KeyError:
			return character, orig, dest

	def iter_unhandled_rules(self, branch, turn, tick):
		portals_with_rulebook_changed = {
			(character, orig, dest)
			for (
				character,
				orig,
				dest,
				_,
				_,
				_,
			) in self.engine._portals_rulebooks_cache.shallowest
		}
		portals_default_rulebooks = {
			(None, port) for port in self.engine._edges_cache.keys
		}
		portals_with_filled_default_rulebooks = {
			(character, orig, dest)
			for (_, (character, orig, dest)) in (
				portals_default_rulebooks
				& self.engine._rulebooks_cache.branches.keys()
			)
		}
		for character, orig, dest in (
			portals_with_rulebook_changed
			| portals_with_filled_default_rulebooks
		):
			try:
				rulebook = self.get_rulebook(
					character, orig, dest, branch, turn, tick
				)
				prio = self.engine._rulebooks_cache.retrieve(
					rulebook, branch, turn, tick
				)[1]
				rules = self.unhandled_rulebook_rules(
					character, orig, dest, rulebook, branch, turn, tick
				)
			except KeyError:
				continue
			for rule in rules:
				yield prio, character, orig, dest, rulebook, rule


class ThingsCache(Cache):
	name = "things_cache"

	def __init__(self, db):
		Cache.__init__(self, db)
		self._make_node = db.thing_cls

	def store(self, *args, planning=None, loading=False, contra=True):
		character, thing, branch, turn, tick, location = args
		with self._lock:
			try:
				oldloc = self.retrieve(character, thing, branch, turn, tick)
			except KeyError:
				oldloc = None
			super().store(
				*args, planning=planning, loading=loading, contra=contra
			)
			node_contents_cache = self.db._node_contents_cache
			this = frozenset((thing,))
			# Cache the contents of nodes
			if oldloc is not None:
				oldconts_orig = node_contents_cache.retrieve(
					character, oldloc, branch, turn, tick
				)
				newconts_orig = oldconts_orig.difference(this)
				node_contents_cache.store(
					character,
					oldloc,
					branch,
					turn,
					tick,
					newconts_orig,
					contra=False,
					loading=True,
				)
				todo = []
				# update any future contents caches pertaining to the old location
				if (character, oldloc) in node_contents_cache.loc_settings:
					locset = node_contents_cache.loc_settings[
						character, oldloc
					][branch]
					if turn in locset:
						for future_tick in locset[turn].future(tick):
							todo.append((turn, future_tick))
					for future_turn, future_ticks in locset.future(
						turn
					).items():
						for future_tick in future_ticks:
							todo.append((future_turn, future_tick))
				for trn, tck in todo:
					node_contents_cache.store(
						character,
						oldloc,
						branch,
						trn,
						tck,
						node_contents_cache.retrieve(
							character, oldloc, branch, trn, tck, search=True
						).difference(this),
						planning=False,
						contra=False,
						loading=True,
					)
			if location is not None:
				todo = []
				try:
					oldconts_dest = node_contents_cache.retrieve(
						character, location, branch, turn, tick
					)
				except KeyError:
					oldconts_dest = frozenset()
				newconts_dest = oldconts_dest.union(this)
				node_contents_cache.store(
					character,
					location,
					branch,
					turn,
					tick,
					newconts_dest,
					contra=False,
					loading=True,
				)
				# and the new location
				if (character, location) in node_contents_cache.loc_settings:
					locset = node_contents_cache.loc_settings[
						character, location
					][branch]
					if turn in locset:
						for future_tick in locset[turn].future(tick):
							todo.append((turn, future_tick))
					for future_turn, future_ticks in locset.future(
						turn
					).items():
						for future_tick in future_ticks:
							todo.append((future_turn, future_tick))
				for trn, tck in todo:
					node_contents_cache.store(
						character,
						location,
						branch,
						trn,
						tck,
						node_contents_cache.retrieve(
							character, location, branch, trn, tck, search=True
						).union(this),
						planning=False,
						contra=False,
						loading=True,
					)

	def turn_before(self, character, thing, branch, turn):
		with self._lock:
			try:
				self.retrieve(character, thing, branch, turn, 0)
			except KeyError:
				pass
			return self.keys[(character,)][thing][branch].rev_before(turn)

	def turn_after(self, character, thing, branch, turn):
		with self._lock:
			try:
				self.retrieve(character, thing, branch, turn, 0)
			except KeyError:
				pass
			return self.keys[(character,)][thing][branch].rev_after(turn)


class NodeContentsCache(Cache):
	name = "node_contents_cache"

	def __init__(self, db, kfkvs=None):
		super().__init__(db, kfkvs)
		self.loc_settings = StructuredDefaultDict(1, SettingsTurnDict)

	def store(
		self,
		character: Key,
		place: Key,
		branch: str,
		turn: int,
		tick: int,
		contents: frozenset,
		planning: bool = None,
		forward: bool = None,
		loading=False,
		contra=None,
	):
		self.loc_settings[character, place][branch].store_at(
			turn, tick, contents
		)

		return super().store(
			character,
			place,
			branch,
			turn,
			tick,
			contents,
			planning=planning,
			forward=forward,
			loading=loading,
			contra=contra,
		)

	def _iter_future_contradictions(
		self, entity, key, turns, branch, turn, tick, value
	):
		return self.db._things_cache._iter_future_contradictions(
			entity, key, turns, branch, turn, tick, value
		)

	def remove(self, branch, turn, tick):
		"""Delete data on or after this tick

		On the assumption that the future has been invalidated.

		"""
		with self._lock:
			assert not self.parents  # not how stuff is stored in this cache
			for branchkey, branches in list(self.branches.items()):
				if branch in branches:
					branhc = branches[branch]
					if turn in branhc:
						trun = branhc[turn]
						if tick in trun:
							del trun[tick]
						trun.truncate(tick)
						if not trun:
							del branhc[turn]
					branhc.truncate(turn)
					if not branhc:
						del branches[branch]
				if not branches:
					del self.branches[branchkey]
			for keykey, keys in list(self.keys.items()):
				for key, branchs in list(keys.items()):
					if branch in branchs:
						branhc = branchs[branch]
						if turn in branhc:
							trun = branhc[turn]
							if tick in trun:
								del trun[tick]
							trun.truncate(tick)
							if not trun:
								del branhc[turn]
						branhc.truncate(turn)
						if not branhc:
							del branchs[branch]
					if not branchs:
						del keys[key]
				if not keys:
					del self.keys[keykey]
			sets = self.settings[branch]
			if turn in sets:
				setsturn = sets[turn]
				if tick in setsturn:
					del setsturn[tick]
				setsturn.truncate(tick)
				if not setsturn:
					del sets[turn]
			sets.truncate(turn)
			if not sets:
				del self.settings[branch]
			presets = self.presettings[branch]
			if turn in presets:
				presetsturn = presets[turn]
				if tick in presetsturn:
					del presetsturn[tick]
				presetsturn.truncate(tick)
				if not presetsturn:
					del presets[turn]
			presets.truncate(turn)
			if not presets:
				del self.presettings[branch]
			for entity, brnch in list(self.keycache):
				if brnch == branch:
					kc = self.keycache[entity, brnch]
					if turn in kc:
						kcturn = kc[turn]
						if tick in kcturn:
							del kcturn[tick]
						kcturn.truncate(tick)
						if not kcturn:
							del kc[turn]
					kc.truncate(turn)
					if not kc:
						del self.keycache[entity, brnch]
			self.shallowest = OrderedDict()



from itertools import product

import networkx as nx

import pytest

from LiSE.character import grid_2d_8graph


@pytest.fixture
def three3(engy):
	return engy.new_character("3x3", grid_2d_8graph(3, 3))


def test_sim_start(three3):
	@three3.place.rule(neighborhood=1, always=True)
	def did_it_run(place):
		if "it_ran" in place:
			place.character.stat["it_ran"] = (
				place.character.stat.get("it_ran", 0) + 1
			)
		else:
			place["it_ran"] = True

	eng = three3.engine
	eng.next_turn()
	assert three3.place[1, 1]["it_ran"]
	assert "it_ran" not in three3.stat
	eng.next_turn()
	assert three3.place[1, 1]["it_ran"]
	assert three3.stat["it_ran"] == 9
	eng.next_turn()
	assert three3.stat["it_ran"] == 9  # still!


@pytest.mark.parametrize(
	("branched", "rulebook"), product(*[(True, False)] * 2)
)
def test_rule_neighborhood(serial_engine, branched, rulebook):
	"""Test a rule applied to all nodes of a character with a neighborhood"""
	char = serial_engine.new_character("char", nx.grid_2d_graph(5, 5))

	@serial_engine.rule
	def it_ran(node):
		node["it_ran"] = True

	@it_ran.trigger
	def should_it_run(node):
		node["trigger_evaluated"] = True
		for neighbor in node.neighbors():
			if neighbor.get("should_run"):
				return True
		return False

	it_ran.neighborhood = 1
	assert it_ran.neighborhood == 1
	if rulebook:
		serial_engine.rulebook["it_ran"] = [it_ran]

	serial_engine.next_turn()
	if rulebook:
		char.place.rulebook = "it_ran"
	else:
		char.place.rule(it_ran)
	char.place[3, 3]["should_run"] = True
	assert it_ran.neighborhood == 1
	if branched:
		serial_engine.branch = "eeeee"
	serial_engine.next_turn()

	for nabor in [(2, 3), (4, 3), (3, 4), (3, 2)]:
		assert char.place[nabor]["it_ran"]

	for outer in [(1, 1), (4, 4)]:
		assert "trigger_evaluated" not in char.place[outer]



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""A simple local server providing access to a LiSE core

Run this as:

	python3 -m LiSE.server

and it will start an HTTP server at localhost:8080. Send msgpack mappings
to it, with the key 'command' set to the name of one of the methods in
:class:`LiSE.handle.EngineHandle` and remaining keys set to arguments
accepted by that method.

Refer to :class:`LiSE.handle.EngineHandle` for documentation on those
methods.

"""

import cherrypy
from argparse import ArgumentParser
from . import LiSEHandleWebService

parser = ArgumentParser()
parser.add_argument("--prefix", action="store", default=".")
args = parser.parse_args()
conf = {
	"/": {
		"request.dispatch": cherrypy.dispatch.MethodDispatcher(),
		"tools.sessions.on": True,
		"tools.response_headers.on": True,
		"tools.response_headers.headers": [
			("Content-Type", "application/json")
		],
		"tools.encode.on": True,
		"tools.encode.encoding": "utf-8",
	}
}
cherrypy.quickstart(LiSEHandleWebService(args.prefix), "/", conf)



# This file is part of allegedb, an object relational mapper for versioned graphs.
# Copyright (C) Zachary Spector. public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Wrapper classes to let you store mutable data types in the allegedb ORM

The wrapper objects act like regular mutable objects, but write a new copy
of themselves to allegedb every time they are changed.

"""

from functools import partial
from itertools import zip_longest
from abc import ABC, abstractmethod
from collections.abc import (
	MutableSet,
	MutableMapping,
	MutableSequence,
	Mapping,
	Sequence,
	Iterable,
	Sized,
	Container,
)
from typing import Callable


class MutableWrapper(ABC):
	__slots__ = ()

	def __iter__(self):
		return iter(self._getter())

	def __len__(self):
		return len(self._getter())

	def __contains__(self, item):
		return item in self._getter()

	def __repr__(self):
		return "<{} instance at {}, wrapping {}>".format(
			self.__class__.__name__, id(self), self._getter()
		)

	def __str__(self):
		return str(self._getter())

	@abstractmethod
	def _getter(self):
		raise NotImplementedError

	@abstractmethod
	def _copy(self):
		raise NotImplementedError

	@abstractmethod
	def _set(self, v):
		raise NotImplementedError

	@abstractmethod
	def unwrap(self):
		raise NotImplementedError


Iterable.register(MutableWrapper)
Sized.register(MutableWrapper)
Container.register(MutableWrapper)


class MutableWrapperDictList(MutableWrapper, ABC):
	__slots__ = ()

	def _subset(self, k, v):
		new = self._copy()
		new[k] = v
		self._set(new)

	def __getitem__(self, k):
		ret = self._getter()[k]
		if isinstance(ret, dict):
			return SubDictWrapper(
				lambda: self._getter()[k], partial(self._subset, k)
			)
		if isinstance(ret, list):
			return SubListWrapper(
				lambda: self._getter()[k], partial(self._subset, k)
			)
		if isinstance(ret, set):
			return SubSetWrapper(
				lambda: self._getter()[k], partial(self._subset, k)
			)
		return ret

	def __setitem__(self, key, value):
		me = self._copy()
		me[key] = value
		self._set(me)

	def __delitem__(self, key):
		me = self._copy()
		del me[key]
		self._set(me)


class MutableMappingUnwrapper(MutableMapping, ABC):
	__slots__ = ()

	def __eq__(self, other):
		if self is other:
			return True
		if not isinstance(other, Mapping):
			return False
		if self.keys() != other.keys():
			return False
		for k in self.keys():
			me = self[k]
			you = other[k]
			if hasattr(me, "unwrap"):
				me = me.unwrap()
			if hasattr(you, "unwrap"):
				you = you.unwrap()
			if me != you:
				return False
		else:
			return True

	def unwrap(self):
		"""Deep copy myself as a dict, all contents unwrapped"""
		return {
			k: v.unwrap()
			if hasattr(v, "unwrap") and not hasattr(v, "no_unwrap")
			else v
			for (k, v) in self.items()
		}


class MutableMappingWrapper(
	MutableWrapperDictList, MutableMappingUnwrapper, ABC
):
	def __eq__(self, other):
		return MutableMappingUnwrapper.__eq__(self, other)

	def unwrap(self):
		return MutableMappingUnwrapper.unwrap(self)


class SubDictWrapper(MutableMappingWrapper, dict):
	__slots__ = ("_getter", "_set")
	_getter: Callable
	_set: Callable

	def __init__(self, getter, setter):
		super().__init__()
		self._getter = getter
		self._set = setter

	def _copy(self):
		return dict(self._getter())

	def _subset(self, k, v):
		new = dict(self._getter())
		new[k] = v
		self._set(new)


class MutableSequenceWrapper(MutableWrapperDictList, MutableSequence, ABC):
	def __eq__(self, other):
		if self is other:
			return True
		if not isinstance(other, Sequence):
			return NotImplemented
		for me, you in zip_longest(self, other):
			if hasattr(me, "unwrap"):
				me = me.unwrap()
			if hasattr(you, "unwrap"):
				you = you.unwrap()
			if me != you:
				return False
		else:
			return True

	def unwrap(self):
		"""Deep copy myself as a list, all contents unwrapped"""
		return [v.unwrap() if hasattr(v, "unwrap") else v for v in self]


class SubListWrapper(MutableSequenceWrapper, list):
	__slots__ = ("_getter", "_set")
	_getter: Callable
	_set: Callable

	def __init__(self, getter, setter):
		super().__init__()
		self._getter = getter
		self._set = setter

	def _copy(self):
		return list(self._getter())

	def insert(self, index, object):
		me = self._copy()
		me.insert(index, object)
		self._set(me)

	def append(self, object):
		me = self._copy()
		me.append(object)
		self._set(me)


class MutableWrapperSet(MutableWrapper, MutableSet):
	__slots__ = ()
	_getter: Callable
	_set: Callable

	def _copy(self):
		return set(self._getter())

	def pop(self):
		me = self._copy()
		yours = me.pop()
		self._set(me)
		return yours

	def discard(self, element):
		me = self._copy()
		me.discard(element)
		self._set(me)

	def remove(self, element):
		me = self._copy()
		me.remove(element)
		self._set(me)

	def add(self, element):
		me = self._copy()
		me.add(element)
		self._set(me)

	def unwrap(self):
		"""Deep copy myself as a set, all contents unwrapped"""
		return {
			v.unwrap()
			if hasattr(v, "unwrap") and not hasattr(v, "no_unwrap")
			else v
			for v in self
		}


class SubSetWrapper(MutableWrapperSet, set):
	__slots__ = ("_getter", "_set")
	_getter: Callable
	_set: Callable

	def __init__(self, getter, setter):
		super().__init__()
		self._getter = getter
		self._set = setter

	def _copy(self):
		return set(self._getter())


class DictWrapper(MutableMappingWrapper, dict):
	"""A dictionary synchronized with a serialized field.

	This is meant to be used in allegedb entities (graph, node, or
	edge), for when the user stores a dictionary in them.

	"""

	__slots__ = ("_getter", "_setter", "_outer", "_key")
	_getter: Callable

	def __init__(self, getter, setter, outer, key):
		super().__init__()
		self._getter = getter
		self._setter = setter
		self._outer = outer
		self._key = key

	def _copy(self):
		return dict(self._getter())

	def _set(self, v):
		self._setter(v)
		self._outer[self._key] = v


class ListWrapper(MutableWrapperDictList, MutableSequence, list):
	"""A list synchronized with a serialized field.

	This is meant to be used in allegedb entities (graph, node, or
	edge), for when the user stores a list in them.

	"""

	__slots__ = ("_getter", "_setter", "_outer", "_key")

	def __init__(self, getter, setter, outer, key):
		self._outer = outer
		self._key = key
		self._getter = getter
		self._setter = setter

	def __eq__(self, other):
		if self is other:
			return True
		if not isinstance(other, Sequence):
			return NotImplemented
		for me, you in zip_longest(self, other):
			if hasattr(me, "unwrap"):
				me = me.unwrap()
			if hasattr(you, "unwrap"):
				you = you.unwrap()
			if me != you:
				return False
		else:
			return True

	def _copy(self):
		return list(self._getter())

	def _set(self, v):
		self._setter(v)
		self._outer[self._key] = v

	def insert(self, i, v):
		new = self._copy()
		new.insert(i, v)
		self._set(new)

	def append(self, v):
		new = self._copy()
		new.append(v)
		self._set(new)

	def unwrap(self):
		"""Deep copy myself as a list, with all contents unwrapped"""
		return [
			v.unwrap()
			if hasattr(v, "unwrap") and not hasattr(v, "no_unwrap")
			else v
			for v in self
		]


class SetWrapper(MutableWrapperSet, set):
	"""A set synchronized with a serialized field.

	This is meant to be used in allegedb entities (graph, node, or
	edge), for when the user stores a set in them.

	"""

	__slots__ = ("_getter", "_setter", "_outer", "_key")
	_getter: Callable

	def __init__(self, getter, setter, outer, key):
		super().__init__()
		self._getter = getter
		self._setter = setter
		self._outer = outer
		self._key = key

	def _set(self, v):
		self._setter(v)
		self._outer[self._key] = v


class UnwrappingDict(dict):
	"""Dict that stores the data from the wrapper classes

	Won't store those objects themselves.

	"""

	def __setitem__(self, key, value):
		if isinstance(value, MutableWrapper):
			value = value.unwrap()
		super(UnwrappingDict, self).__setitem__(key, value)



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The nodes of LiSE's character graphs.

Every node that actually exists is either a Place or a Thing, but they
have a lot in common.

"""

from __future__ import annotations
from collections.abc import Mapping, ValuesView, Set
from typing import Optional, Union, Iterator, List

import networkx as nx
from networkx import shortest_path, shortest_path_length

from .allegedb import graph, Key, HistoricKeyError

from .util import getatt, AbstractCharacter
from .query import StatusAlias
from . import rule
from .exc import AmbiguousUserError, TravelException


class UserMapping(Mapping):
	"""A mapping of the characters that have a particular node as a unit.

	Getting characters from here isn't any better than getting them from
	the engine direct, but with this you can do things like use the
	.get() method to get a character if it's a user and otherwise
	get something else; or test whether the character's name is in
	the keys; and so on.

	"""

	__slots__ = ["node"]

	def __init__(self, node) -> None:
		"""Store the node"""
		self.node = node

	engine = getatt("node.engine")

	def _user_names(self) -> Iterator[Key]:
		node = self.node
		engine = self.engine
		charn = node.character.name
		nn = node.name
		cache = engine._unitness_cache.user_order
		if charn not in cache or nn not in cache[charn]:
			return
		cache = cache[charn][nn]
		seen = set()
		for user in cache:
			if user in seen:
				continue
			for branch, turn, tick in engine._iter_parent_btt():
				if branch in cache[user]:
					branchd = cache[user][branch]
					if turn in branchd:
						if branchd[turn].rev_gettable(tick):
							if branchd[turn][tick]:
								yield user
							seen.add(user)
							break
					elif branchd.rev_gettable(turn):
						turnd = branchd[turn]
						if turnd.final():
							yield user
						seen.add(user)
						break

	@property
	def only(self) -> Node:
		"""If there's only one unit, return it.

		Otherwise, raise ``AmbiguousUserError``, a type of ``AttributeError``.

		"""
		if len(self) != 1:
			raise AmbiguousUserError("No users, or more than one")
		return next(iter(self.values()))

	def __iter__(self) -> Iterator[Key]:
		yield from self._user_names()

	def __len__(self) -> int:
		n = 0
		for user in self._user_names():
			n += 1
		return n

	def __bool__(self) -> bool:
		for user in self._user_names():
			return True
		return False

	def __contains__(self, item) -> bool:
		chara = self.engine.character[item]
		nn = self.node.name
		charn = self.node.character.name
		return charn in chara.unit and nn in chara.unit[charn]

	def __getitem__(self, k) -> AbstractCharacter:
		ret = self.engine.character[k]
		node = self.node
		charn = node.character.name
		nn = node.name
		avatar = ret.unit
		if charn not in avatar or nn not in avatar[charn]:
			raise KeyError("{} not used by {}".format(self.node.name, k))
		return ret


class NodeContentValues(ValuesView):
	_mapping: "NodeContent"

	def __iter__(self) -> Iterator[Thing]:
		node = self._mapping.node
		nodem = node.character.node
		try:
			conts = node.engine._node_contents(node.character.name, node.name)
		except KeyError:
			return
		for name in conts:
			if name not in nodem:
				return
			yield nodem[name]

	def __contains__(self, item) -> bool:
		try:
			return item.location == self._mapping.node
		except AttributeError:
			return False


class NodeContent(Mapping):
	__slots__ = ("node",)

	def __init__(self, node) -> None:
		self.node = node

	def __iter__(self) -> Iterator:
		try:
			it = self.node.engine._node_contents_cache.retrieve(
				self.node.character.name,
				self.node.name,
				*self.node.engine._btt(),
			)
		except KeyError:
			return
		yield from it

	def __len__(self) -> int:
		try:
			return len(
				self.node.engine._node_contents_cache.retrieve(
					self.node.character.name,
					self.node.name,
					*self.node.engine._btt(),
				)
			)
		except KeyError:
			return 0

	def __contains__(self, item) -> bool:
		try:
			return self.node.character.thing[item].location == self.node
		except KeyError:
			return False

	def __getitem__(self, item) -> Thing:
		if item not in self:
			raise KeyError
		return self.node.character.thing[item]

	def values(self) -> NodeContentValues:
		return NodeContentValues(self)


class DestsValues(ValuesView):
	_mapping: "Dests"

	def __contains__(self, item) -> bool:
		_, name = self._mapping._pn
		return item.origin.name == name


class Dests(Mapping):
	__slots__ = ("_ecnb", "_pn")

	def __init__(self, node) -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self) -> Iterator:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.iter_successors(charname, name, *btt())

	def __len__(self) -> int:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.count_successors(charname, name, *btt())

	def __contains__(self, item) -> bool:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_successor(charname, name, item, *btt())

	def __getitem__(self, item) -> "LiSE.portal.Portal":
		portal, name = self._pn
		return portal[name][item]

	def values(self) -> DestsValues:
		return DestsValues(self)


class OrigsValues(ValuesView):
	_mapping: "Origs"

	def __contains__(self, item) -> bool:
		_, name = self._mapping._pn
		return item.destination.name == name


class Origs(Mapping):
	__slots__ = ("_pn", "_ecnb")

	def __init__(self, node) -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self) -> Iterator[Node]:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.iter_predecessors(charname, name, *btt())

	def __contains__(self, item) -> bool:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_predecessor(charname, name, item, *btt())

	def __len__(self) -> int:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.count_predecessors(charname, name, *btt())

	def __getitem__(self, item) -> Node:
		if item not in self:
			raise KeyError
		portal, name = self._pn
		return portal[item][name]

	def values(self) -> OrigsValues:
		return OrigsValues(self)


class Portals(Set):
	__slots__ = ("_pn", "_pecnb")

	def __init__(self, node) -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._pn = (character.portal, name)
		self._pecnb = (
			engine._get_edge,
			engine._edges_cache,
			character,
			character.name,
			name,
			engine._btt,
		)

	def __contains__(self, x) -> bool:
		_, edges_cache, _, charname, name, btt_f = self._pecnb
		btt = btt_f()
		return edges_cache.has_predecessor(
			charname, name, x, *btt
		) or edges_cache.has_successor(charname, name, x, *btt)

	def __len__(self) -> int:
		_, edges_cache, _, charname, name, btt_f = self._pecnb
		btt = btt_f()
		return edges_cache.count_predecessors(
			charname, name, *btt
		) + edges_cache.count_successors(charname, name, *btt)

	def __iter__(self) -> Iterator["LiSE.portal.Portal"]:
		get_edge, edges_cache, character, charname, name, btt_f = self._pecnb
		btt = btt_f()
		for dest in edges_cache.iter_successors(charname, name, *btt):
			yield get_edge(character, name, dest, 0)
		for orig in edges_cache.iter_predecessors(charname, name, *btt):
			yield get_edge(character, orig, name, 0)


class NeighborValues(ValuesView):
	_mapping: "NeighborMapping"

	def __contains__(self, item) -> bool:
		return item.name in self._mapping


class NeighborMapping(Mapping):
	__slots__ = ("_nn", "_ecnb")

	def __init__(self, node: Node) -> None:
		name = node.name
		character = node.character
		engine = node.engine
		self._nn = (character.node, name)
		self._ecnb = (engine._edges_cache, character.name, name, engine._btt)

	def __iter__(self) -> Iterator[Node]:
		edges_cache, charname, name, btt = self._ecnb
		seen = set()
		for succ in edges_cache.iter_successors(charname, name, *btt()):
			yield succ
			seen.add(succ)
		for pred in edges_cache.iter_predecessors(charname, name, *btt()):
			if pred in seen:
				continue
			yield pred
			seen.add(pred)

	def __contains__(self, item) -> bool:
		edges_cache, charname, name, btt = self._ecnb
		return edges_cache.has_predecessor(
			charname, name, item, *btt()
		) or edges_cache.has_successor(charname, name, item, *btt())

	def __len__(self) -> int:
		return len(set(iter(self)))

	def __getitem__(self, item) -> Node:
		node, name = self._nn
		if item not in self:
			raise KeyError(f"{item} is not a neighbor of {name}")
		return node[item]

	def values(self) -> NeighborValues:
		return NeighborValues(self)


class Node(graph.Node, rule.RuleFollower):
	"""The fundamental graph component, which portals go between.

	Every LiSE node is either a thing or a place. They share in common
	the abilities to follow rules; to be connected by portals; and to
	contain things.

	This is truthy if it exists, falsy if it's been deleted.

	"""

	__slots__ = ("_real_rule_mapping",)
	character = getatt("graph")
	name = getatt("node")
	no_unwrap = True
	_extra_keys = {
		"name",
	}

	def _get_rule_mapping(self):
		return rule.RuleMapping(self.db, self.rulebook)

	def _get_rulebook_name(self):
		try:
			return self.engine._nodes_rulebooks_cache.retrieve(
				self.character.name, self.name, *self.engine._btt()
			)
		except KeyError:
			return self.character.name, self.name

	def _get_rulebook(self):
		return rule.RuleBook(self.engine, self._get_rulebook_name())

	def _set_rulebook_name(self, rulebook):
		character = self.character.name
		node = self.name
		cache = self.engine._nodes_rulebooks_cache
		try:
			if rulebook == cache.retrieve(
				character, node, *self.engine._btt()
			):
				return
		except KeyError:
			pass
		branch, turn, tick = self.engine._nbtt()
		cache.store(character, node, branch, turn, tick, rulebook)
		self.engine.query.set_node_rulebook(
			character, node, branch, turn, tick, rulebook
		)

	successor = succ = adj = edge = getatt("portal")
	predecessor = pred = getatt("preportal")
	engine = getatt("db")

	@property
	def user(self) -> UserMapping:
		__doc__ = UserMapping.__doc__
		return UserMapping(self)

	def __init__(self, character, name):
		super().__init__(character, name)
		self.db = character.engine

	@property
	def neighbor(self) -> NeighborMapping:
		return NeighborMapping(self)

	def neighbors(self):
		return self.neighbor.values()

	@property
	def portal(self) -> Dests:
		"""A mapping of portals leading out from this node.

		Aliases ``portal``, ``adj``, ``edge``, ``successor``, and ``succ``
		are available.

		"""
		return Dests(self)

	@property
	def preportal(self) -> Origs:
		"""A mapping of portals leading to this node.

		Aliases ``preportal``, ``predecessor`` and ``pred`` are available.

		"""
		return Origs(self)

	def portals(self) -> Portals:
		"""A set-like object of portals connected to this node."""
		return Portals(self)

	@property
	def content(self) -> NodeContent:
		"""A mapping of ``Thing`` objects that are here"""
		return NodeContent(self)

	def contents(self) -> NodeContentValues:
		"""A set-like object containing ``Thing`` objects that are here"""
		return self.content.values()

	def __iter__(self):
		yield from super().__iter__()
		yield from self._extra_keys
		return

	def clear(self) -> None:
		"""Delete all my keys"""
		for key in super().__iter__():
			del self[key]

	def __contains__(self, k):
		"""Handle extra keys, then delegate."""
		return k in self._extra_keys or super().__contains__(k)

	def __setitem__(self, k, v):
		super().__setitem__(k, v)

	def __delitem__(self, k):
		super().__delitem__(k)

	def successors(self) -> Iterator["Place"]:
		"""Iterate over nodes with edges leading from here to there."""
		for port in self.portal.values():
			yield port.destination

	def predecessors(self) -> Iterator["Place"]:
		"""Iterate over nodes with edges leading here from there."""
		for port in self.preportal.values():
			yield port.origin

	def _plain_dest_name(self, dest):
		if isinstance(dest, Node):
			if dest.character != self.character:
				raise ValueError(
					"{} not in {}".format(dest.name, self.character.name)
				)
			return dest.name
		else:
			if dest in self.character.node:
				return dest
			raise ValueError("{} not in {}".format(dest, self.character.name))

	def shortest_path_length(
		self, dest: Union["Key", "Node"], weight: "Key" = None
	) -> int:
		"""Return the length of the path from me to ``dest``.

		Raise ``ValueError`` if ``dest`` is not a node in my character
		or the name of one.

		"""

		return shortest_path_length(
			self.character, self.name, self._plain_dest_name(dest), weight
		)

	def shortest_path(
		self, dest: Union[Key, "Node"], weight: Key = None
	) -> List[Key]:
		"""Return a list of node names leading from me to ``dest``.

		Raise ``ValueError`` if ``dest`` is not a node in my character
		or the name of one.

		"""
		return shortest_path(
			self.character, self.name, self._plain_dest_name(dest), weight
		)

	def path_exists(
		self, dest: Union[Key, "Node"], weight: Key = None
	) -> bool:
		"""Return whether there is a path leading from me to ``dest``.

		With ``weight``, only consider edges that have a stat by the
		given name.

		Raise ``ValueError`` if ``dest`` is not a node in my character
		or the name of one.

		"""
		try:
			return bool(self.shortest_path_length(dest, weight))
		except KeyError:
			return False

	def delete(self) -> None:
		"""Get rid of this, starting now.

		Apart from deleting the node, this also informs all its users
		that it doesn't exist and therefore can't be their unit
		anymore.

		"""
		self.clear()
		for contained in list(self.contents()):
			contained.delete()
		if self.name in self.character.portal:
			del self.character.portal[self.name]
		if self.name in self.character.preportal:
			del self.character.preportal[self.name]
		for user in list(self.user.values()):
			user.remove_unit(self.character.name, self.name)
		branch, turn, tick = self.engine._nbtt()
		self.engine._nodes_cache.store(
			self.character.name, self.name, branch, turn, tick, False
		)
		self.engine.query.exist_node(
			self.character.name, self.name, branch, turn, tick, False
		)
		self.character.node.send(self.character.node, key=self.name, val=None)

	def new_portal(
		self, other: Union[Key, "Node"], **stats
	) -> "LiSE.portal.Portal":
		"""Connect a portal from here to another node, and return it."""
		return self.character.new_portal(
			self.name, getattr(other, "name", other), **stats
		)

	def new_thing(self, name: Key, **stats) -> "Thing":
		"""Create a new thing, located here, and return it."""
		return self.character.new_thing(name, self.name, **stats)

	def historical(self, stat: Key) -> StatusAlias:
		"""Return a reference to the values that a stat has had in the past.

		You can use the reference in comparisons to make a history
		query, and execute the query by calling it, or passing it to
		``self.engine.ticks_when``.

		"""
		return StatusAlias(entity=self, stat=stat)

	def __bool__(self):
		return self.name in self.character.node


class Place(Node):
	"""The kind of node where a thing might ultimately be located.

	LiSE entities are truthy so long as they exist, falsy if they've
	been deleted.

	"""

	__slots__ = (
		"graph",
		"db",
		"node",
		"_rulebook",
		"_rulebooks",
		"_real_rule_mapping",
	)

	extrakeys = {
		"name",
	}

	def __getitem__(self, key):
		if key == "name":
			return self.name
		return super().__getitem__(key)

	def __repr__(self):
		return "<{}.character[{}].place[{}]>".format(
			repr(self.engine), self.character.name, self.name
		)

	def _validate_node_type(self):
		try:
			self.engine._things_cache.retrieve(
				self.character.name, self.name, *self.engine._btt()
			)
			return False
		except:
			return True

	def delete(self) -> None:
		"""Remove myself from the world model immediately."""
		super().delete()
		self.character.place.send(
			self.character.place, key=self.name, val=None
		)


def roerror(*args):
	raise RuntimeError("Read-only")


class Thing(Node):
	"""The sort of item that has a particular location at any given time.

	Things are always in Places or other Things, and may additionally be
	travelling through a Portal.

	LiSE entities are truthy so long as they exist, falsy if they've
	been deleted.

	"""

	__slots__ = (
		"graph",
		"db",
		"node",
		"_rulebook",
		"_rulebooks",
		"_real_rule_mapping",
	)

	_extra_keys = {"name", "location"}

	def _getname(self):
		return self.name

	def _getloc(self):
		ret = self.engine._things_cache._base_retrieve(
			(self.character.name, self.name, *self.engine._btt())
		)
		if ret is None or isinstance(ret, Exception):
			return None
		return ret

	def _validate_node_type(self):
		return self._getloc() is not None

	def _get_arrival_time(self):
		charn = self.character.name
		n = self.name
		thingcache = self.engine._things_cache
		for b, trn, tck in self.engine._iter_parent_btt():
			try:
				v = thingcache.turn_before(charn, n, b, trn)
			except KeyError:
				v = thingcache.turn_after(charn, n, b, trn)
			if v is not None:
				return v
		else:
			raise ValueError("Couldn't find arrival time")

	def _set_loc(self, loc: Optional[Key]):
		self.engine._set_thing_loc(self.character.name, self.name, loc)

	_getitem_dispatch = {"name": _getname, "location": _getloc}

	_setitem_dispatch = {"name": roerror, "location": _set_loc}

	def __getitem__(self, key: Key):
		"""Return one of my stats stored in the database, or special cases:

		``name``: return the name that uniquely identifies me within
		my Character

		``location``: return the name of my location

		"""
		disp = self._getitem_dispatch
		if key in disp:
			return disp[key](self)
		else:
			return super().__getitem__(key)

	def __setitem__(self, key, value):
		"""Set ``key``=``value`` for the present game-time."""
		try:
			self._setitem_dispatch[key](self, value)
		except HistoricKeyError as ex:
			raise ex
		except KeyError:
			super().__setitem__(key, value)

	def __delitem__(self, key):
		"""As of now, this key isn't mine."""
		if key in self._extra_keys:
			raise ValueError("Can't delete {}".format(key))
		super().__delitem__(key)

	def __repr__(self):
		return "<{}.character['{}'].thing['{}']>".format(
			self.engine, self.character.name, self.name
		)

	def delete(self) -> None:
		super().delete()
		self._set_loc(None)
		self.character.thing.send(
			self.character.thing, key=self.name, val=None
		)

	def clear(self) -> None:
		"""Unset everything."""
		for k in list(self.keys()):
			if k not in self._extra_keys:
				del self[k]

	@property
	def location(self) -> Node:
		"""The ``Thing`` or ``Place`` I'm in."""
		locn = self["location"]
		if locn is None:
			raise AttributeError("Not really a Thing")
		return self.engine._get_node(self.character, locn)

	@location.setter
	def location(self, v: Union[Node, Key]):
		if hasattr(v, "name"):
			v = v.name
		self["location"] = v

	def go_to_place(self, place: Union[Node, Key], weight: Key = None) -> int:
		"""Assuming I'm in a node that has a :class:`Portal` direct
		to the given node, schedule myself to travel to the
		given :class:`Place`, taking an amount of time indicated by
		the ``weight`` stat on the :class:`Portal`, if given; else 1
		turn.

		Return the number of turns the travel will take.

		"""
		if hasattr(place, "name"):
			placen = place.name
		else:
			placen = place
		curloc = self["location"]
		orm = self.character.engine
		turns = (
			1
			if weight is None
			else self.engine._portal_objs[
				(self.character.name, curloc, place)
			].get(weight, 1)
		)
		with self.engine.plan():
			orm.turn += turns
			self["location"] = placen
		return turns

	def follow_path(
		self, path: list, weight: Key = None, check: bool = True
	) -> int:
		"""Go to several nodes in succession, deciding how long to
		spend in each by consulting the ``weight`` stat of the
		:class:`Portal` connecting the one node to the next,
		default 1 turn.

		Return the total number of turns the travel will take. Raise
		:class:`TravelException` if I can't follow the whole path,
		either because some of its nodes don't exist, or because I'm
		scheduled to be somewhere else. Set ``check=False`` if
		you're really sure the path is correct, and this function
		will be faster.

		"""
		if len(path) < 2:
			raise ValueError("Paths need at least 2 nodes")
		eng = self.character.engine
		with eng.world_lock:
			if check:
				prevplace = path.pop(0)
				if prevplace != self["location"]:
					raise ValueError(
						"Path does not start at my present location"
					)
				subpath = [prevplace]
				for place in path:
					if (
						prevplace not in self.character.portal
						or place not in self.character.portal[prevplace]
					):
						raise TravelException(
							"Couldn't follow portal from {} to {}".format(
								prevplace, place
							),
							path=subpath,
							traveller=self,
						)
					subpath.append(place)
					prevplace = place
			else:
				subpath = path.copy()
			turns_total = 0
			prevsubplace = subpath.pop(0)
			turn_incs = []
			myname = self.name
			charn = self.character.name
			branch, turn, tick = eng._btt()
			for subplace in subpath:
				if weight is not None:
					turn_incs.append(
						self.engine._edge_val_cache.retrieve(
							charn,
							prevsubplace,
							subplace,
							0,
							branch,
							turn,
							tick,
						)
					)
				else:
					turn_incs.append(1)
				turns_total += turn_incs[-1]
				turn += turn_incs[-1]
				tick = eng._turn_end_plan.get(turn, 0)
				if check:
					eng._nodes_cache.retrieve(
						charn, subplace, branch, turn, tick
					)
				_, start_turn, start_tick, end_turn, end_tick = eng._branches[
					branch
				]
				if (
					(start_turn < turn < end_turn)
					or (
						start_turn == end_turn == turn
						and start_tick <= tick < end_tick
					)
					or (start_turn == turn and start_tick <= tick)
					or (end_turn == turn and tick < end_tick)
				):
					eng.load_at(branch, turn, tick)
			with eng.plan():
				for subplace, turn_inc in zip(subpath, turn_incs):
					eng.turn += turn_inc
					branch, turn, tick = eng._nbtt()
					eng._things_cache.store(
						charn, myname, branch, turn, tick, subplace
					)
					eng.query.set_thing_loc(
						charn, myname, branch, turn, tick, subplace
					)
			return turns_total

	def travel_to(
		self,
		dest: Union[Node, Key],
		weight: Key = None,
		graph: nx.DiGraph = None,
	) -> int:
		"""Find the shortest path to the given node from where I am
		now, and follow it.

		If supplied, the ``weight`` stat of each :class:`Portal` along
		the path will be used in pathfinding, and for deciding how
		long to stay in each Place along the way.

		The ``graph`` argument may be any NetworkX-style graph. It
		will be used for pathfinding if supplied, otherwise I'll use
		my :class:`Character`. In either case, however, I will attempt
		to actually follow the path using my :class:`Character`, which
		might not be possible if the supplied ``graph`` and my
		:class:`Character` are too different. If it's not possible,
		I'll raise a :class:`TravelException`, whose ``subpath``
		attribute holds the part of the path that I *can* follow. To
		make me follow it, pass it to my ``follow_path`` method.

		Return value is the number of turns the travel will take.

		"""
		destn = dest.name if hasattr(dest, "name") else dest
		if destn == self.location.name:
			raise ValueError("I'm already at {}".format(destn))
		graph = self.character if graph is None else graph
		path = nx.shortest_path(graph, self["location"], destn, weight)
		return self.follow_path(path, weight)



from LiSE import Engine
from LiSE.proxy import EngineProcessManager


def test_serialize_character(engy):
	char = engy.new_character("physical")
	assert engy.unpack(engy.pack(char)) == char


def test_serialize_thing(engy):
	char = engy.new_character("physical")
	place = char.new_place("here")
	thing = place.new_thing("that")
	assert engy.unpack(engy.pack(thing)) == thing


def test_serialize_place(engy):
	char = engy.new_character("physical")
	place = char.new_place("here")
	assert engy.unpack(engy.pack(place)) == place


def test_serialize_portal(engy):
	char = engy.new_character("physical")
	a = char.new_place("a")
	b = char.new_place("b")
	port = a.new_portal(b)
	assert engy.unpack(engy.pack(port)) == port


def test_serialize_function(tmp_path):
	with Engine(
		tmp_path, random_seed=69105, enforce_end_of_time=False, workers=0
	) as eng:

		@eng.function
		def foo(bar: str, bas: str) -> str:
			return bar + bas + " is correct"

	procm = EngineProcessManager()
	engprox = procm.start(tmp_path)
	funcprox = engprox.function.foo
	assert funcprox("foo", "bar") == "foobar is correct"
	procm.shutdown()


def test_serialize_method(tmp_path):
	with Engine(
		tmp_path, random_seed=69105, enforce_end_of_time=False, workers=0
	) as eng:

		@eng.method
		def foo(self, bar: str, bas: str) -> str:
			return bar + bas + " is correct"

	procm = EngineProcessManager()
	engprox = procm.start(tmp_path)
	assert engprox.foo("bar", "bas") == "barbas is correct"
	procm.shutdown()



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import pytest

from LiSE.allegedb import OutOfTimelineError


@pytest.fixture(scope="function")
def chara(engy):
	yield engy.new_character("chara")


def test_many_things_in_place(chara):
	place = chara.new_place(0)
	things = [place.new_thing(i) for i in range(1, 10)]
	for thing in things:
		assert thing in place.contents()
	for that in place.content:
		assert place.content[that].location == place
	things.sort(key=lambda th: th.name)
	contents = sorted(place.contents(), key=lambda th: th.name)
	assert things == contents


def test_contents_over_time(chara):
	place = chara.new_place(0)
	correct_contents = set()
	for i in range(10):
		chara.engine.next_turn()
		with chara.engine.advancing():
			place.new_thing(chara.engine.turn)
			del chara.thing[chara.engine.turn]
			assert set(place.content.keys()) == correct_contents
			place.new_thing(chara.engine.turn)
			correct_contents.add(chara.engine.turn)
		assert set(place.content.keys()) == correct_contents
	del chara.thing[9]
	correct_contents.remove(9)
	assert set(place.content.keys()) == correct_contents
	del chara.thing[8]
	correct_contents.remove(8)
	assert set(place.content.keys()) == correct_contents
	chara.engine.turn = 5
	with pytest.raises(OutOfTimelineError):
		del chara.thing[5]
	chara.engine.branch = "b"
	del chara.thing[5]
	assert set(place.content.keys()) == {1, 2, 3, 4}


def test_contents_in_plan(chara):
	place = chara.new_place(0)
	correct_contents = {1, 2, 3, 4, 5}
	for th in correct_contents:
		place.new_thing(th)
	with chara.engine.plan():
		for i in range(6, 15):
			chara.engine.turn += 1
			assert set(place.content) == correct_contents
			place.new_thing(i)
			del chara.thing[i]
			assert set(place.content) == correct_contents
			place.new_thing(i)
			correct_contents.add(i)
			assert set(place.content) == correct_contents
	chara.engine.turn = 4
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7, 8, 9}
	chara.engine.turn = 2
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7}
	# this does not contradict the plan
	place.new_thing(15)
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7, 15}
	chara.engine.turn = 4
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7, 8, 9, 15}
	# this neither
	there = chara.new_place("there")
	chara.thing[9].location = there
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7, 8, 15}
	chara.engine.turn = 5
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7, 8, 10, 15}
	chara.engine.turn = 10
	assert set(place.content) == correct_contents.union({15}).difference({9})
	# but this does
	chara.engine.turn = 5
	place.new_thing(11)
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 15}
	chara.engine.turn = 10
	assert set(place.content) == {1, 2, 3, 4, 5, 6, 7, 8, 10, 11, 15}



# This file is part of LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Directed edges, as used by LiSE."""

from __future__ import annotations
from collections.abc import Mapping
from typing import Union, List, Tuple, Any

from .allegedb.graph import Edge
from .allegedb import HistoricKeyError, Key

from .util import getatt, AbstractCharacter
from .query import StatusAlias
from .rule import RuleFollower
from .rule import RuleMapping as BaseRuleMapping


class RuleMapping(BaseRuleMapping):
	"""Mapping to get rules followed by a portal."""

	def __init__(self, portal):
		"""Store portal, engine, and rulebook."""
		super().__init__(portal.engine, portal.rulebook)
		self.portal = portal


class Portal(Edge, RuleFollower):
	"""Connection between two nodes that :class:`LiSE.node.Thing` travel along

	LiSE entities are truthy so long as they exist, falsy if they've
	been deleted.

	"""

	__slots__ = (
		"graph",
		"orig",
		"dest",
		"idx",
		"origin",
		"destination",
		"_rulebook",
		"_real_rule_mapping",
	)
	character = getatt("graph")
	engine = getatt("db")
	no_unwrap = True

	def __init__(self, graph: AbstractCharacter, orig: Key, dest: Key):
		super().__init__(graph, orig, dest, 0)
		self.origin = graph.node[orig]
		self.destination = graph.node[dest]

	@property
	def _cache(self):
		return self.db._edge_val_cache[self.character.name][self.orig][
			self.dest
		][0]

	def _rule_name_activeness(self):
		rulebook_name = self._get_rulebook_name()
		cache = self.engine._active_rules_cache
		if rulebook_name not in cache:
			return
		cache = cache[rulebook_name]
		for rule in cache:
			for branch, turn, tick in self.engine._iter_parent_btt():
				if branch not in cache[rule]:
					continue
				try:
					yield (rule, cache[rule][branch][turn][tick])
					break
				except ValueError:
					continue
				except HistoricKeyError as ex:
					if ex.deleted:
						break
		raise KeyError("{}->{} has no rulebook?".format(self.orig, self.dest))

	def _get_rulebook_name(self):
		try:
			return self.engine._portals_rulebooks_cache.retrieve(
				self.character.name, self.orig, self.dest, *self.engine._btt()
			)
		except KeyError:
			return (self.character.name, self.orig, self.dest)

	def _set_rulebook_name(self, rulebook):
		character = self.character.name
		orig = self.orig
		dest = self.dest
		cache = self.engine._portals_rulebooks_cache
		try:
			if rulebook == cache.retrieve(
				character, orig, dest, *self.engine._btt()
			):
				return
		except KeyError:
			pass
		branch, turn, tick = self.engine._nbtt()
		cache.store(character, orig, dest, branch, turn, tick, rulebook)
		self.engine.query.set_portal_rulebook(
			character, orig, dest, branch, turn, tick, rulebook
		)

	def _get_rule_mapping(self):
		return RuleMapping(self)

	def __getitem__(self, key):
		if key == "origin":
			return self.orig
		elif key == "destination":
			return self.dest
		elif key == "character":
			return self.character.name
		else:
			return super().__getitem__(key)

	def __setitem__(self, key, value):
		if key in ("origin", "destination", "character"):
			raise KeyError("Can't change " + key)
		super().__setitem__(key, value)

	def __repr__(self):
		"""Describe character, origin, and destination"""
		return "<{}.character[{}].portal[{}][{}]>".format(
			repr(self.engine),
			repr(self["character"]),
			repr(self["origin"]),
			repr(self["destination"]),
		)

	def __bool__(self):
		"""It means something that I exist, even if I have no data."""
		return (
			self.orig in self.character.portal
			and self.dest in self.character.portal[self.orig]
		)

	@property
	def reciprocal(self) -> "Portal":
		"""If there's another Portal connecting the same origin and
		destination that I do, but going the opposite way, return
		it. Else raise KeyError.

		"""
		try:
			return self.character.portal[self.dest][self.orig]
		except KeyError:
			raise AttributeError("This portal has no reciprocal")

	def historical(self, stat: Key) -> StatusAlias:
		"""Return a reference to the values that a stat has had in the past.

		You can use the reference in comparisons to make a history
		query, and execute the query by calling it, or passing it to
		``self.engine.ticks_when``.

		"""
		return StatusAlias(entity=self, stat=stat)

	def update(
		self, e: Union[Mapping, List[Tuple[Any, Any]]] = None, **f
	) -> None:
		"""Works like regular update, but less

		Only actually updates when the new value and the old value differ.
		This is necessary to prevent certain infinite loops.

		"""
		if e is not None:
			if hasattr(e, "keys") and callable(e.keys):
				for k in e.keys():
					if k not in self:
						self[k] = e[k]
					else:
						v = e[k]
						if self[k] != v:
							self[k] = v
			else:
				for k, v in e:
					if k not in self or self[k] != v:
						self[k] = v
		for k, v in f.items():
			if k not in self or self[k] != v:
				self[k] = v

	def delete(self) -> None:
		"""Remove myself from my :class:`Character`.

		For symmetry with :class:`Thing` and :class:`Place`.

		"""
		self.clear()
		self.engine._exist_edge(
			self.character.name, self.orig, self.dest, exist=None
		)

	def unwrap(self) -> dict:
		"""Return a dictionary representation of this entity"""
		return {
			k: v.unwrap()
			if hasattr(v, "unwrap") and not hasattr(v, "no_unwrap")
			else v
			for (k, v) in self.items()
		}
